import contextlib
import io


def import_this(**kwargs):
    """Print the Zen of Python"""
    # https://stackoverflow.com/a/23794519
    zen = io.StringIO()
    with contextlib.redirect_stdout(zen):
        import this  # noqa: F401

    text = f"```{zen.getvalue()}```"
    return text



"""
0. Save/copy this file to a new file under commands/

1. Add logic to my_command and rename it to something more meaningful.
   Optionally you can access user, channel, text from the passed in **kwargs (don't remove this).

2. Add a useful docstring to your renamed my_command.

3. Return a message string that the command/bot should post to the channel.
   You probably want to add some code under __main__ to make sure the function does what you want ...

4. In bot/slack.py import the script, e.g.: from commands.template import my_command

5. Add the command to the appropriate dict: ADMIN_BOT_COMMANDS, PUBLIC_BOT_COMMANDS or PRIVATE_BOT_COMMANDS
   (public = for us in channels, private = for use in @karmabot DM)

6. PR your work. Thanks
"""


def my_command(**kwargs: dict) -> str:  # type: ignore
    """Text that will appear in the help section"""
    # kwargs will hold user, channel, text (from a Slack message object)

    # use them like this, or just delete these line:
    user_id = kwargs.get("user_id")  # noqa: F841
    channel_id = kwargs.get("channel")  # noqa: F841
    msg_text = kwargs.get("text")  # noqa: F841

    # return a message string
    msg = "Replace me!"
    return msg


if __name__ == "__main__":
    #  standalone test
    user, channel, text = "ABCD123", "XYZ789", "some message"
    kwargs = dict(user=user, channel=channel, text=text)
    output = my_command(**kwargs)  # type: ignore
    print(output)



import datetime
import re
from typing import Callable, Dict, Tuple, Union

from karmabot.db.database import database
from karmabot.db.karma_note import KarmaNote
from karmabot.db.karma_user import KarmaUser

NOTE_CMD_PATTERN = re.compile(r"note\s(\w+)\s?(.*)")


def note(user_id: str, channel: str, text: str) -> Union[None, str]:
    """Allows the user to store and retrieve simple notes.

    - Syntax for adding a note: @karmabot note add <">my note<"> (note message can be in quotes)
    - Syntax for listing notes: @karmabot note list
    - Syntax for removing a note: @karmabote note del 1

    Each note is stored for the current user only. A user can only list and delete her own notes.
    """
    user_id = user_id.strip("<>@")

    # retrieve current user
    with database.session_manager() as session:
        user = session.query(KarmaUser).get(user_id)

    cmd, _ = _parse_note_cmd(text)

    note_cmd_fnc = NOTE_COMMANDS.get(cmd, _command_not_found)

    return note_cmd_fnc(text, user)


def _add_note(text: str, user: KarmaUser) -> str:
    """Adds a new note to the database for the given user."""
    _, note_msg = _parse_note_cmd(text)
    if not note_msg:
        return f"Sorry {user.username}, could not find a note in your message."

    if _note_exists(note_msg, user):
        return f"Sorry {user.username}, you already have an identical note."

    note = KarmaNote(
        user_id=user.user_id, timestamp=datetime.datetime.now(), note=note_msg
    )

    with database.session_manager() as session:
        session.add(note)
        session.commit()

    return f"Hey {user.username}, you've just stored a note."


def _del_note(text: str, user: KarmaUser) -> str:
    """Deletes the note with the given note id."""
    _, note_id = _parse_note_cmd(text)

    if not note_id:
        return f"Sorry {user.username}, it seems you did not provide a valid id."

    with database.session_manager() as session:
        query = session.query(KarmaNote).filter_by(id=note_id, user_id=user.user_id)

        row_count = query.delete()
        session.commit()  # otherwise, the deletion is not performed

    if row_count:
        return f"Hey {user.username}, your note was successfully deleted."

    return (
        f"Sorry {user.username}, something went wrong, no record was deleted. "
        f"Please ask an admin..."
    )


def _list_notes(text: str, user: KarmaUser) -> str:
    """List all notes for a given user."""
    notes = _get_notes_for_user(user)

    if not notes:
        return (
            f"Sorry {user.username}, you don't have any notes so far! "
            f"Just start adding notes via the 'note add' command."
        )

    msg = "\n".join(f"{i+1}. note {str(note)}" for i, note in enumerate(notes))

    return msg


def _command_not_found(text: str, user: KarmaUser) -> str:
    return (
        f"Sorry {user.username}, your note command was not recognized. "
        f"You can use {', '.join(NOTE_COMMANDS.keys())}."
    )


def _parse_note_cmd(text: str) -> Tuple[str, str]:
    note_cmd = ("", "")

    match = NOTE_CMD_PATTERN.search(text)
    if match:
        note_cmd = match.group(1).strip(), match.group(2).strip("\"'")

    return note_cmd


def _get_notes_for_user(user: KarmaUser) -> list:
    with database.session_manager() as session:
        notes = session.query(KarmaNote).filter_by(user_id=user.user_id).all()
    return notes


def _note_exists(msg: str, user: KarmaUser) -> bool:
    with database.session_manager() as session:
        selected_note = session.query(KarmaNote).filter_by(
            note=msg, user_id=user.user_id
        )
        note_exists = session.query(
            selected_note.exists()
        ).scalar()  # returns True or False
    return note_exists


NOTE_COMMANDS: Dict[str, Callable] = {
    "add": _add_note,
    "del": _del_note,
    "list": _list_notes,
}



import logging
import os
import re
from pathlib import Path

from dotenv import load_dotenv


def get_env_var(env_var: str, default: str = None) -> str:
    env_var_value = os.environ.get(env_var)

    # explicit check for None as None is returned by environ.get for non existing keys
    if env_var_value is None:
        if default is not None:
            return default

        raise KeyError(
            f"{env_var} was not found. Please check your .karmabot file as well as the README.md."
        )

    if not env_var_value:
        raise ValueError(
            f"{env_var} was found but seems empty. Please check your .karmabot file as well as the README.md."
        )

    return env_var_value


dotenv_path = Path(".").resolve() / ".karmabot"
if dotenv_path.exists():
    logging.info("Loading environmental variables from: '%s'", dotenv_path)
    load_dotenv(dotenv_path)

# Environment variables
KARMABOT_ID = get_env_var("KARMABOT_SLACK_USER")
DATABASE_URL = get_env_var("KARMABOT_DATABASE_URL")
SLACK_APP_TOKEN = get_env_var("KARMABOT_SLACK_APP_TOKEN")
SLACK_BOT_TOKEN = get_env_var("KARMABOT_SLACK_BOT_TOKEN")
TEST_MODE = bool(get_env_var("KARMABOT_TEST_MODE") == "true")
logging.info("Test mode enabled: %s", TEST_MODE)

# Slack
GENERAL_CHANNEL = get_env_var("KARMABOT_GENERAL_CHANNEL")
ADMINS = get_env_var("KARMABOT_ADMINS")
ADMINS = ADMINS.split(",")  # type: ignore

# Karma
# the first +/- is merely signaling, start counting (regex capture)
# from second +/- onwards, so bob++ adds 1 point, bob+++ = +2, etc
KARMA_ACTION_PATTERN = re.compile(
    r"(?:^| )(\S{2,}?)\s?[\+\-]([\+\-]+)", flags=re.MULTILINE
)
MAX_POINTS = 5



import random

import requests

PLATFORM = "https://codechalleng.es"
CC_ES = "CodeChalleng.es"
TIPS_ENDPOINT = f"{PLATFORM}/api/tips"
NEW_TIP_LINK = f"{PLATFORM}/inbox/new/pytip/"

ADD_TIP = f"\nSource: {CC_ES} | Share more tips: {NEW_TIP_LINK}\n"


def get_random_tip(**kwargs):
    """Print a random Python tip, quote or nugget from CodeChalleng.es"""
    resp = requests.get(TIPS_ENDPOINT)
    tips = resp.json()
    tip = random.choice(tips)
    msg = f"> {tip['tip']}\n"

    if tip["link"]:
        msg += f"\n{tip['link']}\n"

    if tip["code"]:
        msg += f"\n```{tip['code']}```\n"

    msg += ADD_TIP
    return msg


if __name__ == "__main__":
    tip = get_random_tip()
    print(tip)



import ssl

import feedparser

# https://stackoverflow.com/a/28296087
if hasattr(ssl, "_create_unverified_context"):
    ssl._create_default_https_context = ssl._create_unverified_context

RSS = "https://pybit.es/feeds/all.rss.xml"
MAX_ENTRIES = 5


def get_pybites_last_entries(**kwargs):
    """Get the last 5 entries of PyBites blog (might take a bit)"""
    data = feedparser.parse(RSS)

    output = []
    for item in data["entries"][:MAX_ENTRIES]:
        title = item["title"]
        published = item["published"]
        url = item["link"]
        output.append(f"{title} ({published})\n{url}\n")

    return "\n".join(output)


if __name__ == "__main__":
    output = get_pybites_last_entries()
    print(output)



import logging
from typing import Dict, List

import karmabot.bot as bot
from karmabot.slack import get_user_id


def join_public_channels(**kwargs):
    """Makes the bot join all public channels he is not in yet"""
    response = bot.app.client.conversations_list(
        exclude_archived=True, types="public_channel"
    )
    if response["ok"]:
        public_channels = response["channels"]
    else:
        error = "Could not retrieve public channel list"
        logging.error(error)
        return error

    channels_joined = []
    for channel in public_channels:
        if not channel["is_member"]:  # only join channels you are not already in
            bot.app.client.conversations_join(channel=channel["id"])
            bot.app.client.chat_postMessage(
                text="Karmabot is here! :wave:", channel=channel["id"]
            )
            channels_joined.append(channel["name"])

    if channels_joined:
        channels_text = ",".join(channels_joined)
        return f"I joined the following public channels: {channels_text}"

    return "There were no new public channels to join!"


def your_id(**kwargs):
    """Shows the user id of karmabot"""
    message = kwargs.get("text", "")
    bot_slack_id = message.split()[0]
    bot_user_id = get_user_id(bot_slack_id)

    if bot_user_id:
        return f"My user id is: {bot_user_id}"

    return "Sorry could not retrieve my user id"


def general_channel_id(**kwargs):
    """Shows the channel id of the general channel"""
    response: Dict = bot.app.client.conversations_list(
        exclude_archived=True, types="public_channel"
    )

    if not response["ok"]:
        logging.error('Error for API call "channels.list": %s', response["error"])
        return "I am truly sorry but something went wrong ;("

    channels: List[Dict] = response["channels"]
    for channel in channels:
        if channel["is_general"]:
            general_id = channel["id"]
            return f"The general channel id is: {general_id}"

    return "Sorry, could not find the general channel id :("



import logging
import sys
from contextlib import contextmanager

import sqlalchemy as sa
import sqlalchemy.orm as orm
from sqlalchemy.exc import OperationalError

from karmabot import settings
from karmabot.db.modelbase import SqlAlchemyBase
from karmabot.exceptions import NotInitializedException


class Database:
    """Class for handlading the database connection and provide access to sessions"""

    def __init__(
        self, connection_string: str = settings.DATABASE_URL, echo: bool = False
    ) -> None:

        self.connection_string = connection_string
        self.echo = echo

        self._engine = None
        self._SessionFactory = None  # noqa

    def connect(self):
        """Sets up connection to DB and initializes models"""
        logging.debug("Connecting to DB with %s", self.connection_string)

        self._engine = sa.create_engine(self.connection_string, echo=self.echo)
        try:
            self._engine.connect()
        except OperationalError:
            logging.error("Database connection failed.")
            sys.exit(1)

        logging.info("DB connection successful")
        self._SessionFactory = orm.sessionmaker(bind=self.engine)

    def create_models(self):
        import karmabot.db.__all_models  # noqa: F401

        if not self._engine:
            raise NotInitializedException("SqlAlchemy engine is not initialized")
        SqlAlchemyBase.metadata.create_all(self._engine)

    @contextmanager
    def session_manager(self):
        if not self._SessionFactory:
            raise NotInitializedException("SessionFactory is not initialized")

        session = self._SessionFactory()

        try:
            yield session
        except Exception:
            logging.error("Rollback database transaction")
            session.rollback()
            raise
        finally:
            logging.debug("Closing database connection")
            session.close()

    @property
    def session(self):
        if not self._SessionFactory:
            raise NotInitializedException("SessionFactory is not initialized")
        return self._SessionFactory()

    @property
    def engine(self):
        if not self._engine:
            raise NotInitializedException("SqlAlchemy engine is not initialized")
        return self._engine


database = Database()



import logging

import karmabot.bot as bot
import karmabot.slack
from karmabot.db.database import database
from karmabot.db.karma_user import KarmaUser


def update_username(**kwargs):
    """Changes the Username"""
    user_id = kwargs.get("user_id")
    if user_id:
        user_id = user_id.strip("<>@")

    with database.session_manager() as session:
        karma_user: KarmaUser = session.query(KarmaUser).get(user_id)

    if not karma_user:
        return "User not found"

    old_username = karma_user.username

    response = bot.app.client.users_profile_get(user=user_id)
    status = response.status_code

    if status != 200:
        logging.error("Cannot get user info for %s - API error: %s", user_id, status)
        return "Sorry, I could not retrieve your user information from the slack API :("

    user_profile = response.data["profile"]
    new_username = karmabot.slack.get_available_username(user_profile)

    if old_username == new_username:
        return (
            f"Sorry, you have not updated your username: {old_username}. \n"
            "Please update your real-name or display-name in your Slack "
            "profile and retry."
        )

    with database.session_manager() as session:
        karma_user.username = new_username
        session.commit()

    return (
        f"Sucessfully updated your KarmaUser name "
        f"from '{old_username}' to '{new_username}'!"
    )


def get_user_name(**kwargs):
    """Shows the current Username"""
    user_id = kwargs.get("user_id")
    if user_id:
        user_id = user_id.strip("<>@")

    with database.session_manager() as session:
        karma_user: KarmaUser = session.query(KarmaUser).get(user_id)

    if not karma_user:
        return "Sorry, you are not yet known to karmabot. Try to give some Karma! :)"
    username = karma_user.username

    return f"Your current username for karmabot is '{username}'"



""" private command, not callable """
from random import choice

import karmabot.slack
from karmabot.settings import KARMABOT_ID, get_env_var

# thanks Erik!
WELCOME_MSG = """Welcome {user} ++!

Introduce yourself in #general if you like ...
- What do you use Python for?
- What is your day job?
- And: >>> random.choice(pybites_init_questions)
{welcome_question}

Although you will meet some awesome folks here, you can also talk to me :)
Type `help` here to get started ...

Enjoy PyBites Slack and keep calm and code in Python!

{admins}
"""
# some Pythonic welcome questions
WELCOME_QUESTIONS = """How did you use Python for the first time?
What is your favorite Python module?
What was the last Python book you read?
Did you go to Pycon? If so what was the best you got out of it?
Do you have any particular interest or hobby?
How did you hear about PyBites?
What is your favorite software principle of the Zen of Python (import this)
Are you a 100 Days of Code survivor or planning to take the challenge?
What is your favorite TV show, movie or book?
How many Christopher Nolan movies did you see? If > 1 favorite?
If you were to build a chatbot what would it do?
AI, hype or life threatening?
How do you currently use Python?
Are you teaching Python or inspire to do so?
Australia or Spain?
Star Trek or Star Wars?
Tabs or spaces? (be careful!)
Do you use test drive development (TDD)?
What is your favorite editor?
What other programming languages do you know and/or use?"""


def _get_admins() -> str:
    admins_env = get_env_var("KARMABOT_ADMINS", default=KARMABOT_ID).split(",")
    admins = ", ".join(f"<@{admin}>" for admin in admins_env)
    return " and".join(admins.rsplit(",", maxsplit=1))


def welcome_user(user_id: str) -> str:
    """Welcome a new PyBites community member"""
    questions = WELCOME_QUESTIONS.split("\n")
    random_question = choice(questions)  # noqa: S311
    slack_id = karmabot.slack.get_slack_id(user_id)
    admins = _get_admins()
    return WELCOME_MSG.format(
        user=slack_id, welcome_question=random_question, admins=admins
    )



import sqlalchemy as sa

from karmabot.db.modelbase import SqlAlchemyBase


class KarmaUser(SqlAlchemyBase):
    """Models a slack user with karma in the DB"""

    __tablename__ = "karma_user"

    user_id = sa.Column(sa.String, primary_key=True)
    username = sa.Column(sa.String)
    karma_points = sa.Column(sa.Integer, default=0)

    def formatted_user_id(self):
        """Formats user id for use in slack messages"""
        return f"<@{self.user_id}>"

    def __repr__(self):
        return (
            f"<KarmaUser> ID: {self.user_id} | Username: {self.username} | "
            f"Karma-Points: {self.karma_points}"
        )



import logging

import karmabot.bot as bot
import karmabot.slack as slack
from karmabot.db.database import database
from karmabot.db.karma_transaction import KarmaTransaction
from karmabot.db.karma_user import KarmaUser
from karmabot.exceptions import GetUserInfoException
from karmabot.settings import KARMABOT_ID, MAX_POINTS


class Karma:
    def __init__(self, giver_id, receiver_id, channel_id):
        self.session = database.session
        self.giver: KarmaUser = self.session.query(KarmaUser).get(giver_id)
        self.receiver: KarmaUser = self.session.query(KarmaUser).get(receiver_id)

        if not self.giver:
            self.giver = self._create_karma_user(giver_id)
        if not self.receiver:
            self.receiver = self._create_karma_user(receiver_id)

        self.channel_id: str = channel_id
        self.last_score_maxed_out: bool = False

    def _create_karma_user(self, user_id):
        response = bot.app.client.users_profile_get(user=user_id)
        status = response.status_code

        if status != 200:
            logging.info("Cannot get user info for %s - API error: %s", user_id, status)
            raise GetUserInfoException

        user_profile = response.data["profile"]
        username = slack.get_available_username(user_profile)

        new_user = KarmaUser(user_id=user_id, username=username)
        self.session.add(new_user)
        self.session.commit()

        logging.info("Created new KarmaUser: %s", repr(new_user))
        return new_user

    def _calc_final_score(self, points):
        if abs(points) > MAX_POINTS:
            self.last_score_maxed_out = True
            return MAX_POINTS if points > 0 else -MAX_POINTS
        else:
            self.last_score_maxed_out = False
            return points

    def _create_msg_bot_self_karma(self, points) -> str:
        if points > 0:
            text = (
                f"Thanks {self.giver.username} for the extra karma"
                f", my karma is {self.receiver.karma_points} now"
            )
        else:
            text = (
                f"Not cool {self.giver.username} lowering my karma "
                f"to {self.receiver.karma_points}, but you are probably"
                f" right, I will work harder next time"
            )
        return text

    def _create_msg(self, points):
        receiver_name = self.receiver.username

        poses = "'" if receiver_name.endswith("s") else "'s"
        action = "increase" if points > 0 else "decrease"

        text = (
            f"{receiver_name}{poses} karma {action}d to "
            f"{self.receiver.karma_points}"
        )
        if self.last_score_maxed_out:
            text += f" (= max {action} of {MAX_POINTS})"

        return text

    def _save_transaction(self, points):

        response = bot.app.client.conversations_info(channel=self.channel_id)

        if response.status_code != 200:
            raise Exception(
                f"Slack API could not get Channel info - Status {response.status_code}"
            )

        channel_name = response.data["channel"]["name"]

        transaction = KarmaTransaction(
            giver_id=self.giver.user_id,
            receiver_id=self.receiver.user_id,
            channel=channel_name,
            karma=points,
        )

        self.session.add(transaction)
        self.session.commit()

        finished_transaction = (
            self.session.query(KarmaTransaction)
            .order_by(KarmaTransaction.id.desc())
            .first()
        )
        logging.info(repr(finished_transaction))

    def change_karma(self, points):
        """Updates Karma in the database"""
        if not isinstance(points, int):
            err = "change_karma should not be called with a non int points arg!"
            raise RuntimeError(err)

        try:
            if self.receiver.user_id == self.giver.user_id:
                raise ValueError("Sorry, cannot give karma to self")

            points = self._calc_final_score(points)
            self.receiver.karma_points += points
            self.session.commit()

            self._save_transaction(points)

            if self.receiver.user_id == KARMABOT_ID:
                return self._create_msg_bot_self_karma(points)
            else:
                return self._create_msg(points)

        finally:
            logging.info(
                "[Karmachange] %s to %s: %s",
                self.giver.user_id,
                self.receiver.user_id,
                points,
            )
            self.session.close()


def _parse_karma_change(karma_change):
    user_id, voting = karma_change

    receiver = slack.get_user_id(user_id)
    points = voting.count("+") - voting.count("-")

    return receiver, points


def process_karma_changes(karma_giver, channel_id, karma_changes):
    messages = []
    for karma_change in karma_changes:
        receiver_id, points = _parse_karma_change(karma_change)
        try:
            karma = Karma(
                giver_id=karma_giver,
                receiver_id=receiver_id,
                channel_id=channel_id,
            )
        except GetUserInfoException:
            return "Sorry, something went wrong while retrieving user information"

        try:
            text = karma.change_karma(points)
        except (RuntimeError, ValueError) as exc:
            text = str(exc)

        messages.append(text)

    return messages



"""The PyBites Karmabot for Slack."""
try:
    from importlib.metadata import PackageNotFoundError, version  # type: ignore
except ImportError:  # pragma: no cover
    from importlib_metadata import PackageNotFoundError, version  # type: ignore


try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"


import logging

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(name)-12s %(levelname)-8s %(message)s",  # noqa E501
    datefmt="%m-%d %H:%M",
    handlers=[logging.StreamHandler()],
)

logging.info("Karmabot started")



import datetime

import sqlalchemy as sa

from karmabot.db.modelbase import SqlAlchemyBase


class KarmaNote(SqlAlchemyBase):
    """Models a simple note system in the DB"""

    __tablename__ = "karma_note"

    id: int = sa.Column(  # noqa
        sa.BigInteger().with_variant(sa.Integer, "sqlite"),
        primary_key=True,
        autoincrement=True,
    )
    user_id = sa.Column(sa.String, sa.ForeignKey("karma_user.user_id"), nullable=False)
    timestamp = sa.Column(sa.DateTime, default=datetime.datetime.now, nullable=False)
    note = sa.Column(sa.String)

    def __repr__(self):
        return (
            f"[KarmaNote] ID: {self.id} | {self.user_id} -> "
            f"{self.timestamp} | Note: {self.note}"
        )

    def __str__(self):
        return (
            f"(ID: {self.id}) from {self.timestamp.strftime('%Y-%m-%d, %H:%M')}: "
            f"{self.note}."
        )



import karmabot.slack

MSG = """Hey {username}, so you want to propose a new command eh?

Awesome! Here are the steps:
1. Karmabot repo: https://github.com/pybites/karmabot
2. Fork the repo, make your branch.
3. Add your command script under the commands subdirectory.
4. Open a PR of your branch against PyBites repo.
5. Bob/Julian/Community to approve and merge it in.
6. Join the slack channel: #karmabot_dev

Here is a walk-through video:
https://www.youtube.com/watch?v=Yx9qYl6lmzM&t=2s

Thanks!
"""


def add_command(**kwargs):
    """Instructions how to propose a new bot command"""
    user_id = kwargs.get("user_id")
    if not user_id:
        return None

    slack_id = karmabot.slack.get_slack_id(user_id)
    return MSG.format(username=slack_id)



import logging
import re
import unicodedata
from typing import Callable, Dict

from slack_bolt import App

# Commands
from karmabot.commands.add import add_command
from karmabot.commands.age import pybites_age
from karmabot.commands.control import general_channel_id, join_public_channels, your_id
from karmabot.commands.doc import doc_command
from karmabot.commands.feed import get_pybites_last_entries
from karmabot.commands.joke import joke
from karmabot.commands.note import note
from karmabot.commands.score import get_karma, top_karma
from karmabot.commands.tip import get_random_tip
from karmabot.commands.topchannels import get_recommended_channels
from karmabot.commands.update_username import get_user_name, update_username
from karmabot.commands.welcome import welcome_user
from karmabot.commands.zen import import_this
from karmabot.exceptions import CommandExecutionException
from karmabot.karma import process_karma_changes

# Settings
from karmabot.settings import (
    ADMINS,
    KARMA_ACTION_PATTERN,
    KARMABOT_ID,
    SLACK_BOT_TOKEN,
    TEST_MODE,
)
from karmabot.slack import MessageChannelType

# command constants
ADMIN_BOT_COMMANDS = {
    "top_karma": top_karma,
    "join_public_channels": join_public_channels,
    "your_id": your_id,
    "general_channel_id": general_channel_id,
}
CHANNEL_BOT_COMMANDS = {
    "add": add_command,
    "age": pybites_age,
    "joke": joke,
    "note": note,
    "tip": get_random_tip,
    "topchannels": get_recommended_channels,
    "zen": import_this,
}
DM_BOT_COMMANDS = {
    "doc": doc_command,
    "feed": get_pybites_last_entries,
    "joke": joke,
    "note": note,
    "karma": get_karma,
    "topchannels": get_recommended_channels,
    "username": get_user_name,
    "updateusername": update_username,
}

# other constants
SPECIAL_REPLIES = {
    "zen": "`import this`",
    "cheers": ":beers:",
    "braces": "`SyntaxError: not a chance`",
}


def compile_command_pattern(commands: Dict[str, Callable]) -> re.Pattern:
    command_words = commands.keys()
    all_commands = "|".join(command_words)

    full_commands = fr"^<@{KARMABOT_ID}>\s*({all_commands})(\s.*)?$"

    return re.compile(full_commands, re.IGNORECASE)


def compile_special_reply_pattern(replies: Dict[str, str]) -> re.Pattern:
    special_words = "|".join(replies.keys())
    pattern = fr"(?<!<@{KARMABOT_ID}>\s)({special_words})"
    return re.compile(pattern, flags=re.MULTILINE | re.IGNORECASE)


ADMIN_COMMAND_PATTERN = compile_command_pattern(ADMIN_BOT_COMMANDS)
CHANNEL_COMMAND_PATTERN = compile_command_pattern(CHANNEL_BOT_COMMANDS)
DM_COMMAND_PATTERN = compile_command_pattern(DM_BOT_COMMANDS)
UNKNOWN_COMMAND_PATTERN = re.compile(fr"^<@{KARMABOT_ID}>\s(\w*)")
HELP_COMMAND_PATTERN = re.compile(fr"^<@{KARMABOT_ID}>\s(help|commands)")
SPECIAL_WORDS_PATTERN = compile_special_reply_pattern(SPECIAL_REPLIES)
COMMAND_ERROR = "Sorry, something went wrong when performing the requested command"

# Slack Bolt App Init
# For testing we disable the token_verification, such that no valied token is required
TOKEN_VERIFICATION = not TEST_MODE
app = App(
    token=SLACK_BOT_TOKEN,
    name="Karmabot",
    token_verification_enabled=TOKEN_VERIFICATION,
)  # type: ignore


# Helpers
def perform_command(commands, cmd_match, params):
    cmd = cmd_match.group(1)
    try:
        return commands[cmd](**params)
    except KeyError:
        return f"No command '{cmd}' found"


def create_commands_table(commands):
    """Print this help text"""
    ret = "\n".join(
        [
            f"{name:<30}: {func.__doc__.strip()}"
            for name, func in sorted(commands.items())
        ]
    )
    return f"```{ret}```"


# Top priority: process karma
@app.message(KARMA_ACTION_PATTERN)  # type: ignore
def karma_action(message, say):
    msg = unicodedata.normalize("NFKD", message["text"])

    karma_giver = message["user"]
    channel_id = message["channel"]
    karma_changes = KARMA_ACTION_PATTERN.findall(msg)

    karma_replies = process_karma_changes(karma_giver, channel_id, karma_changes)

    reply = "\n\n".join(karma_replies)
    say(reply)


# Help
@app.message(HELP_COMMAND_PATTERN)  # type: ignore
def reply_help(message, say):
    """Sends the list of available commands as DM to the user"""
    user_id = message["user"]
    channel_type = message["channel_type"]

    help_msg = [
        "\n1. Channel commands (format: `@karmabot command`)",
        create_commands_table(CHANNEL_BOT_COMMANDS),
        "\n2. Message commands (type `@karmabot command` in a DM to the bot)",
        create_commands_table(DM_BOT_COMMANDS),
    ]

    if user_id in ADMINS and channel_type == MessageChannelType.DM.value:
        help_msg.append("\n3. Admin only commands")
        help_msg.append(create_commands_table(ADMIN_BOT_COMMANDS))

    text = "\n".join(help_msg)
    say(text=text, channel=user_id)


# Message replies
@app.message(SPECIAL_WORDS_PATTERN)  # type: ignore
def reply_special_words(message, say):
    msg = message["text"].lower()
    special_word = SPECIAL_WORDS_PATTERN.findall(msg)[0]
    special_reply = SPECIAL_REPLIES.get(special_word)

    text = f"To _{special_word}_ I say: {special_reply}"

    say(text)


# Commands
@app.event("message")  # type: ignore
def reply_commands(message, say):  # noqa
    """
    Handles all the commands in one place

    Unfortunatly we cannot create sperate functions for every category
    (admin, private public) as some commands fit in multiple catagories and the
    first matching function would "swallow" the message and not forwared
    it further down the line
    """
    try:
        user_id = message["user"]
        channel_id = message["channel"]
        text = message["text"]
        channel_type = message["channel_type"]
    except KeyError as exc:
        logging.error("reply_commands error! Message was: %s", message, exc_info=exc)
        return

    kwargs = {"user_id": user_id, "channel": channel_id, "text": text}
    cmd_result = None

    admin_match = ADMIN_COMMAND_PATTERN.match(text)
    if (
        admin_match
        and channel_type == MessageChannelType.DM.value
        and user_id in ADMINS
    ):
        cmd_result = perform_command(ADMIN_BOT_COMMANDS, admin_match, kwargs)
        if not cmd_result:
            say(COMMAND_ERROR)
            raise CommandExecutionException(text)

    private_match = DM_COMMAND_PATTERN.match(text)
    if private_match and channel_type == MessageChannelType.DM.value:
        cmd_result = perform_command(DM_BOT_COMMANDS, private_match, kwargs)
        if not cmd_result:
            say(COMMAND_ERROR)
            raise CommandExecutionException(text)

    public_match = CHANNEL_COMMAND_PATTERN.match(text)
    if public_match and channel_type in [
        MessageChannelType.CHANNEL.value,
        MessageChannelType.GROUP.value,
    ]:
        cmd_result = perform_command(CHANNEL_BOT_COMMANDS, public_match, kwargs)
        if not cmd_result:
            say(COMMAND_ERROR)
            raise CommandExecutionException(text)

    if cmd_result:  # reply with result to a valid cmd
        say(cmd_result)
    elif UNKNOWN_COMMAND_PATTERN.match(text):  # everything else that looks like a cmd
        unknown_cmd = UNKNOWN_COMMAND_PATTERN.findall(text)[0]
        say(f'Sorry <@{user_id}>, there is no command "{unknown_cmd}"')

    # all other messages just do not get a reply


# Events
@app.event("team_join")  # type: ignore
def welcome_new_user(event, say):
    user_id = event["user"]["id"]
    text = welcome_user(user_id)
    logging.info("Sending welcome DM to new member %s", user_id)
    say(text=text, channel=user_id)


@app.event("channel_created")  # type: ignore
def autojoin_new_channels(event, say):
    new_channel_id = event["channel"]["id"]
    app.client.conversations_join(channel=new_channel_id)

    text = (
        "Awesome, a new PyBites channel! Birds of a feather flock together! "
        "Keep doing your nerdy stuff, I will keep track of your karmas :)"
    )
    say(text=text, channel=new_channel_id)



import difflib
from typing import Union

import pyjokes

import karmabot.slack

PYJOKE_HREF = "<https://pyjok.es/|PyJoke>"
CATEGORIES = list(pyjokes.jokes_en.jokes_en.keys())


def joke(**kwargs) -> Union[str, None]:
    """Posts a random PyJoke in the public or private channel"""
    user_id, text = kwargs.get("user_id"), kwargs.get("text")

    if user_id is not None and text is not None:
        slack_id = karmabot.slack.get_slack_id(user_id)
        words = text.lower().split()

        user_category = words[-1] if len(words) > 2 else ""
        category = _get_closest_category(user_category)
    else:
        return None

    joke_text = pyjokes.get_joke(category=category)

    return f"Hey {slack_id}, here is a {PYJOKE_HREF} for you: _{joke_text}_"


def _get_closest_category(input_category: str):
    category = difflib.get_close_matches(input_category, CATEGORIES)
    category = category[0] if category else "all"

    return category



class GetUserInfoException(Exception):
    """Excpetion if Slack API UserInfo could net retrieved"""

    pass


class CommandNotFoundException(Exception):
    """Exception if a request Command is not found"""

    pass


class CommandExecutionException(Exception):
    """Exception for failures during command execution"""

    pass


class NotInitializedException(Exception):
    """Exception if required values are not initialized properly"""

    pass



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['karmabot', 'karmabot.commands', 'karmabot.db']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.20,<2.0.0',
 'feedparser>=6.0.8,<7.0.0',
 'freezegun>=1.1.0,<2.0.0',
 'humanize>=3.10.0,<4.0.0',
 'importlib-metadata>=3.7.3,<4.0.0',
 'psycopg2-binary>=2.9.1,<3.0.0',
 'pyjokes>=0.6.0,<0.7.0',
 'python-dotenv>=0.18.0,<0.19.0',
 'requests>=2.25.1,<3.0.0',
 'slack-bolt>=1.6.1,<2.0.0',
 'slack-sdk>=3.7.0,<4.0.0']

entry_points = \
{'console_scripts': ['karmabot = karmabot.main:main']}

setup_kwargs = {
    'name': 'karmabot',
    'version': '2.0.1',
    'description': 'PyBites Karmabot - A Python based Slack Chatbot for Community interaction',
    'long_description': '# PyBites Karmabot - A Python based Slack Chatbot\n\n[![Tests](https://github.com/PyBites-Open-Source/karmabot/workflows/Tests/badge.svg)](https://github.com/PyBites-Open-Source/karmabot/actions?workflow=Tests) [![Black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black) [![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/pre-commit/pre-commit) [![codecov.io](https://codecov.io/github/PyBites-Open-Source/karmabot/coverage.svg?branch=master)](https://codecov.io/github/PyBites-Open-Source/karmabot?branch=master)\n\n**A Python based Slack Chatbot for Community interaction**\n\n## Features\n\nKarmabot\'s main features is the management of Karma within the slack community server. You can give karma, reduce karma, check your current karma points and manage your karma related username.\n\n![karma example](https://www.pogross.de/uploads/karmabot.png)\n\n[Demo Youtube Video](https://www.youtube.com/watch?v=Yx9qYl6lmzM&amp;t=2s)\n\nAdditional commands / features are:\n\n- Jokes powered by [PyJokes](https://github.com/pyjokes/pyjokes)\n- Overview on top channels of the slack server\n- Random Python tip, quote or nugget from CodeChalleng.es\n- Browse and search python documentation, "pydoc help"\n\n## Installation\n\n`pip install karmabot`\n\n## Basic Usage\n\nAfter installing you can start karmabot by using the command\n\n```bash\nkarmabot\n```\n\nHowever, you need to some setup and supply some settings prior to this.\n\n### Setup\n\nFor app creation and tokens please follow the [slack-bolt guide](https://slack.dev/bolt-python/tutorial/getting-started) and enable [socket mode](https://slack.dev/bolt-python/concepts#socket-mode).\n\n#### Settings\n\nBy default we will look for a `.karmabot` file in the directory you used the `karmabot` command. The file should supply the following information.\n\n```env\n# Slack bot app\nKARMABOT_SLACK_BOT_TOKEN=\nKARMABOT_SLACK_APP_TOKEN=\n\n# Workspace\nKARMABOT_SLACK_USER=\nKARMABOT_GENERAL_CHANNEL=\nKARMABOT_ADMINS=\n\n# Backend\nKARMABOT_DATABASE_URL=\n\n# Testing\nKARMABOT_TEST_MODE=\n```\n\nKARMABOT_SLACK_BOT_TOKEN\n:   The [SLACK_BOT_TOKEN](https://slack.dev/bolt-python/tutorial/getting-started) for your bot. You will find it under **OAuth & Permission 🠊 Bot User OAuth Access Token** in your [app](https://api.slack.com/apps/). The token starts with `xoxb-`.\n\nKARMABOT_SLACK_APP_TOKEN\n: The SLACK_APP_TOKEN used for running the bot in [Socket Mode](https://slack.dev/bolt-python/concepts#socket-mode). You will find it under **Basic Information 🠊 App-Level Tokens** in your [app](https://api.slack.com/apps/).\n  The token starts with `xapp-`.\n\nKARMABOT_SLACK_USER\n: The bot\'s user id. Initially, you can fill in a placeholder. Once you\'ve run your own Karmabot for the first time, you can ask it as admin in private chat via `@Karmabot your_id`. This will return a value starting with `U`, e.g., `U0123XYZ`. Replace your placeholder with this value.\n\nKARMABOT_GENERAL_CHANNEL\n: The channel id of your main channel in slack. Initially, you can fill in a placeholder. Once you\'ve run your own Karmabot for the first time, you can ask it as admin in private chat via `@Karmabot general_channel_id`. This will return a value starting with `C`, e.g., `C0123XYZ`. Replace your placeholder with this value.\n\nKARMABOT_ADMINS\n: The [slack user ids](https://api.slack.com/methods/users.identity) of the users that should have admin command access separated by commas.\n\nKARMABOT_DATABASE_URL\n  : The database url which should be compatible with SqlAlchemy. For the provided docker file use `postgresql://user42:pw42@localhost:5432/karmabot`.\n  :heavy_exclamation_mark: To start the provided Docker-based Postgres server, be sure you have Docker Compose [installed](https://docs.docker.com/compose/install/) and run `docker-compose up -d` from the karmabot directory.\n\nKARMABOT_TEST_MODE=\n  : Determines if the code is run in test mode. User `KARMABOT_TEST_MODE=true` to enable testing mode. Everything else will default to `false`. This setting has to be provided as `true`, if you want run tests without a valid `KARMABOT_SLACK_BOT_TOKEN`. Otherwise, you will receive an exceptions with `slack_bolt.error.BoltError: token is invalid ...`.\n\nIf you do not want to use a file you have to provide environment variables with the above names. If no file is present we default to environment variables.\n\n#### Permissions\n\nGo to your [slack app](https://api.slack.com/apps/) and click on **Add features and functionality**. Then go into the following categories and set permissions.\n\n- Event Subscriptions\n  - Enable Events 🠊 Toggle the slider to on\n  - Subscribe to bot events 🠊 Add via the **Add Bot User Event** button\n    - team_join\n    - channel_create\n    - message.channels\n    - message.groups\n    - message.im\n- Permissions\n  - Scopes 🠊 Add the following permissions via the **Add an OAuth Scope** button\n    - app_mentions:read\n    - channels:history\n    - channels:join\n    - channels:read\n    - chat:write\n    - groups:history\n    - groups:read\n    - groups:write\n    - im:history\n    - im:read\n    - im:write\n    - users.profile:read\n    - users:read\n\n## Development pattern for contributors\n\nWe use [poetry](https://github.com/python-poetry/poetry) and `pyproject.toml` for managing packages, dependencies and some settings.\n\n### Setup virtual environment for development\n\nYou should follow the [instructions](https://github.com/python-poetry/poetry) to get poetry up and running for your system. We recommend to use a UNIX-based development system (Linux, Mac, WSL). After setting up poetry you can use `poetry install` within the project folder to install all dependencies.\n\nThe poetry virtual environment should be available in the the project folder as `.venv` folder as specified in `poetry.toml`. This helps with `.venv` detection in IDEs.\n\n#### Conda users\n\nIf you use the Anaconda Python distribution (strongly recommended for Windows users) and `conda create` for your virtual environments, then you will not be able to use the `.venv` environment created by poetry because it is not a conda environment. If you want to use `poetry` disable poetry\'s behavior of creating a new virtual environment with the following command: `poetry config virtualenvs.create false`. You can add `--local` if you don\'t want to change this setting globally but only for the current project. See the [poetry configuration docs](https://python-poetry.org/docs/configuration/) for more details.\n\nNow, when you run `poetry install`, poetry will install all dependencies to your conda environment. You can verify this by running `pip freeze` after `poetry install`.\n\n### Testing and linting\n\nFor testing you need to install [nox](https://nox.thea.codes/en/stable/) separately from the project venv created by poetry. For testing just use the `nox` command within the project folder. You can run all the nox sessions separately if need, e.g.,\n\n- only linting `nox -rs lint`\n- only testing `nox -rs tests`\n\nIf `nox` cannot be found, use `python -m nox` instead.\n\nFor different sessions see the `nox.py` file. You can run `nox --list` to see a list of all available sessions.\n\nIf you want to run tests locally via `pytest` you have to provide a valid `.karmabot` settings file or the respective enviroment variables.\n\nPlease make sure all tests and checks pass before opening pull requests!\n\n#### Using nox under Windows and Linux (WSL)\n\nMake sure to delete the `.nox` folder when you switch from Windows to WSL and vice versa, because the environments are not compatible.\n\n### [pre-commit](https://pre-commit.com/)\n\nTo ensure consistency you can use pre-commit. `pip install pre-commit` and after cloning the karmabot repo run `pre-commit install` within the project folder.\n\nThis will enable pre-commit hooks for checking before every commit.\n',
    'author': 'PyBites',
    'author_email': 'info@pybit.es',
    'maintainer': 'Patrick-Oliver Groß',
    'maintainer_email': 'mail@pogross.de',
    'url': 'https://github.com/PyBites-Open-Source/karmabot',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import datetime

import sqlalchemy as sa
from sqlalchemy import Integer

from karmabot.db.modelbase import SqlAlchemyBase


class KarmaTransaction(SqlAlchemyBase):
    """Models a karma transaction in the DB"""

    __tablename__ = "karma_transaction"

    id: int = sa.Column(  # noqa
        sa.BigInteger().with_variant(Integer, "sqlite"),
        primary_key=True,
        autoincrement=True,
    )
    giver_id = sa.Column(sa.String, sa.ForeignKey("karma_user.user_id"), nullable=False)
    receiver_id = sa.Column(
        sa.String, sa.ForeignKey("karma_user.user_id"), nullable=False
    )

    # data
    timestamp = sa.Column(sa.DateTime, default=datetime.datetime.now, nullable=False)
    channel = sa.Column(sa.String)
    karma = sa.Column(sa.Integer, nullable=False)

    __table_args__ = (sa.CheckConstraint(giver_id != receiver_id),)

    def __repr__(self):
        return (
            f"[KarmaTransaction] ID: {self.id} | {self.giver_id} -> "
            f"{self.receiver_id} | {self.timestamp} | Karma: {self.karma}"
        )



import logging
from collections import namedtuple
from datetime import datetime as dt
from math import exp
from operator import itemgetter
from typing import Dict, List, Optional, Union

import humanize
from slack_sdk.errors import SlackApiError

import karmabot.bot as bot
from karmabot.settings import KARMABOT_ID

MSG_BEGIN = "Glad you asked, here are some channels our Community recommends (based on member count and activity):\n"
MSG_LINE = (
    "- #{channel} ({member_count} members, last post {time_since_last_post}): {purpose}"
)
DEFAULT_NR_CHANNELS = 7

Channel = namedtuple("Channel", "id name purpose num_members latest_ts latest_subtype")


def channel_is_potential(channel):
    is_channel = channel["is_channel"]
    is_member = channel["is_member"]
    is_general = channel["is_general"]
    is_private = channel["is_private"]
    return is_channel and is_member and not is_general and not is_private


def collect_channel_info(channel):
    channel_id = channel["id"]
    try:
        info_response: Dict = bot.app.client.conversations_info(
            channel=channel_id, include_num_members=True
        )
        if not info_response["ok"]:
            raise SlackApiError("converstation.info error", info_response)

        history_response: Dict = bot.app.client.conversations_history(
            channel=channel_id, limit=1
        )
        if not history_response["ok"]:
            raise SlackApiError("conversation.history error", history_response)

    except SlackApiError as exc:
        logging.error(exc)
        return "I am truly sorry but something went wrong ;("

    channel_info: Dict = info_response["channel"]
    channel_history: Dict = history_response["messages"][0]

    latest_ts = channel_history.get("ts")
    latest_type = channel_history.get("type")
    if latest_ts:
        info = Channel(
            channel["id"],
            channel["name"],
            channel_info["purpose"]["value"],
            channel_info["num_members"],
            float(latest_ts),
            latest_type,
        )
        return info

    return None


def get_recommended_channels(**kwargs):
    """Show some of our Community's favorite channels you can join"""
    text = kwargs.get("text")

    potential_channels: Channel = []
    msg = MSG_BEGIN

    if not text:
        nr_channels = DEFAULT_NR_CHANNELS
    else:
        nr_channels = text.split()[2] if len(text.split()) >= 3 else DEFAULT_NR_CHANNELS

    if isinstance(nr_channels, str):
        nr_channels = (
            int(nr_channels) if nr_channels.isnumeric() else DEFAULT_NR_CHANNELS
        )

    # retrieve channel list
    response: Dict = bot.app.client.conversations_list(
        exclude_archived=True, types="public_channel"
    )

    if not response["ok"]:
        logging.error('Error for API call "channels.list": %s', response["error"])
        return "I am truly sorry but something went wrong ;("

    channels: List[Dict] = response["channels"]

    # retrieve channel info for each channel in channel list
    # only consider channels that are not the general channel, that are not private and that have at least one message
    for channel in channels:
        if channel_is_potential(channel):
            info = collect_channel_info(channel)
            if info:
                potential_channels.append(info)

    # now weight channels and return message
    potential_channels = sorted(
        ((calc_channel_score(chan), chan) for chan in potential_channels), reverse=True
    )

    msg = MSG_BEGIN + "\n".join(
        (
            MSG_LINE.format(
                channel=channel.name,
                member_count=channel.num_members,
                time_since_last_post=humanize.naturaltime(
                    seconds_since_last_post(channel)
                ),
                purpose=channel.purpose
                or "<Invest today and get an awesome description!>",
            )
            for score, channel in potential_channels[:nr_channels]
            if score > 0
        )
    )

    return msg


def get_messages(
    channel: Channel, ignore_message_types: Union[set, None] = None
) -> Union[List[Dict], None]:
    """Return a list of the most recent messages in a given channel, filtering
    out certain message types.

    Similar to invite permissions, this requires a user token.

    "New" bot tokens will be both more granular and more flexible, so should
    be able to replace the need for user tokens going forward:

    Ref: https://api.slack.com/docs/token-types#bot_new
    """
    if ignore_message_types is None:
        ignore_message_types = {"channel_join"}

    response = bot.app.client.conversations_history(
        channel=channel.id, user=KARMABOT_ID
    )

    return response["messages"]


def calc_channel_score(channel: Channel):
    """simple calculation of a channels value
    the higher the number of members and the less the number of seconds since the last post the higher the channels score
    """
    since_latest = seconds_since_last_post(channel)
    if not since_latest:
        return 0
    num_members = channel.num_members
    time_delta_in_hours = since_latest / 3600

    return num_members * (exp(-time_delta_in_hours))


def seconds_since_last_post(channel: Channel) -> Optional[float]:
    """return the fraction of days since the last post in a channel, or None if
    all messages are of filtered/ignored subtypes
    """

    ignore_message_types = {"channel_join"}

    if channel.latest_subtype not in ignore_message_types:
        latest_ts = channel.latest_ts
    else:
        msgs = get_messages(channel, ignore_message_types)
        if not msgs:
            return None
        latest_ts = float(max(msgs, key=itemgetter("ts"))["ts"])

    return (dt.now() - dt.fromtimestamp(latest_ts)).total_seconds()



"""A Karmabot pydoc interface.
"""
import contextlib
import io
import pydoc

import karmabot.slack

MSG_APOLOGY = """Sorry {username}, I got nothing for "{text}".

I'll do a keyword search for "{text}" if you add -k before {text}.

Try "topics" or "modules" for more general help.
"""

MSG_FOUNDIT = """Good news {username}, I found the following about {text}:
```
{result}
```
"""


MSG_HELP = """
pydoc [-k keyword] [module_path_or_topic|topics|modules|help]

You can use pydoc to look up all sorts of pythonic things!

Use this to get the docstring for a module:

    pydoc list

Or do a keyword search to get a list of modules that match:

    pydoc -k keyword

Get a list of modules:

    pydoc modules

A list of python language topics, super interesting:

    pydoc topics

And information about the specific listed topics:

    pydoc LOOPING

"""


def doc_command(**kwargs) -> str:
    """Browse and search python documentation, "pydoc help" """
    user_id = str(kwargs.get("user_id"))
    text = str(kwargs.get("text"))

    if len(text) == 0 or text.lower() == "help":
        return MSG_HELP

    apropos = "-k" in text

    if "-" in text and not apropos:  # weed out switches that aren't -k
        return MSG_HELP

    output = io.StringIO()
    with contextlib.redirect_stdout(output):
        if apropos:
            pydoc.apropos(text.partition("-k")[-1])
        else:
            help(text)
    result = output.getvalue()

    slack_id = karmabot.slack.get_slack_id(user_id)

    if result.startswith("No"):
        return MSG_APOLOGY.format(username=slack_id, text=text)

    return MSG_FOUNDIT.format(username=slack_id, text=text, result=result)


if __name__ == "__main__":
    import sys

    kwargs = {"user": "Erik", "channel": "#unix", "text": " ".join(sys.argv[1:])}
    print(doc_command(**kwargs))



import karmabot.slack
from karmabot.db.database import database
from karmabot.db.karma_user import KarmaUser

TOP_NUMBER = 10


def get_karma(**kwargs):
    """Get your current karma score"""
    user_id = kwargs.get("user_id")
    slack_id = karmabot.slack.get_slack_id(user_id)

    with database.session_manager() as session:
        kama_user = session.query(KarmaUser).get(user_id)

    if not kama_user:
        return "User not found"

    if kama_user.karma_points == 0:
        return "Sorry, you don't have any karma yet"

    return f"Hey {slack_id}, your current karma is {kama_user.karma_points}"


def top_karma(**kwargs):
    """Get the PyBites members with most karma"""
    output = ["PyBites members with most karma:"]

    with database.session_manager() as session:
        top_users = (
            session.query(KarmaUser)
            .order_by(KarmaUser.karma_points.desc())
            .limit(TOP_NUMBER)
        )

    if top_users:
        for top_user in top_users:
            output.append(f"{top_user.username:<20} -> {top_user.karma_points}")
        ret = "\n".join(output)
        return f"```{ret}```"

    return "Sorry, no users found"



from datetime import datetime

PYBITES_BORN = datetime(year=2016, month=12, day=19)


def pybites_age(**kwargs) -> str:
    """Print PyBites age in days"""
    today = datetime.now()
    days_old = (today - PYBITES_BORN).days
    return f"PyBites is {days_old} days old"



from slack_bolt.adapter.socket_mode import SocketModeHandler

from karmabot.bot import app
from karmabot.db.database import database
from karmabot.settings import SLACK_APP_TOKEN


def main():
    database.connect()
    database.create_models()

    handler = SocketModeHandler(app, SLACK_APP_TOKEN)
    handler.start()


if __name__ == "__main__":
    main()



import re
from enum import Enum

SLACK_ID_PATTERN = re.compile(r"^<@[^>]+>$")


class MessageChannelType(Enum):
    DM = "im"
    CHANNEL = "channel"
    GROUP = "group"


def get_slack_id(user_id: str) -> str:
    """
    Formats a plain user_id (ABC123XYZ) to use slack identity
    Slack format <@ABC123XYZ> for highlighting users

    :param user_id: Plain user id
    :type user_id: str
    :return: Slack formatted user_id
    :rtype: str
    """
    if SLACK_ID_PATTERN.match(user_id):
        return user_id

    return f"<@{user_id}>"


def get_user_id(user_id: str) -> str:
    """
    Formats the user_id to a plain format removing any <, < or @
    :param user_id: slack id format
    :type user_id: str
    :return: plain user id
    :rtype: str
    """
    return user_id.strip("<>@")


def get_available_username(user_profile):
    """
    Determines the username based on information available from slack.
    First information is used in the following order:
    1) display_name, 2) real_name

    :param user_profile: Slack user_profile dict
    :return: human-readable username
    """

    display_name = user_profile["display_name_normalized"]
    if display_name:
        return display_name

    real_name = user_profile["real_name_normalized"]
    if real_name:
        return real_name

    raise ValueError("User Profile data missing name information")
